//	キャラクタLCD/HD44780互換コントローラ ライブラリ
//	『昼夜逆転』工作室	@jsdiy	http://jsdiy.webcrow.jp
//	2020/04

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "HD44780_lcdlib.h"

//データバスのマスク
#define	LCD_DATABUS	(0x0F << LCD_DB_SHIFT)

//LCD表示情報
typedef	struct
{
	uint8_t XColumnMax, YRowMax;
	uint8_t isCombine, isOverwrite;
	uint8_t cursorPosX, cursorPosY;
}
TLcdInfo;
//
static	TLcdInfo	lcdInfo;

//関数
static	void	LcdCmd4bit(uint8_t data);
static	void	LcdData4bit(uint8_t data);


//下位4bitを逆順に並び替える
#if LCD_DB_REVERSE
static	uint8_t	BitReverse(uint8_t data)
{
	return						//abcd
		((data >> 3) & 0x01) |	//000a |
		((data >> 1) & 0x02) |	//00b0 |
		((data << 1) & 0x04) |	//0c00 |
		((data << 3) & 0x08);	//d000;
}
#endif

//LCDにコマンドを出力する
void	LcdCmd(uint8_t data)
{
	LcdCmd4bit((data >> 4) & 0x0F);	//上位4bit
	LcdCmd4bit(data & 0x0F);		//下位4bit
	
	if (data == LCD_CMD_CLEAR_DISPLAY || data == LCD_CMD_RETURN_HOME )
	{
		lcdInfo.cursorPosX = lcdInfo.cursorPosY = 0;
	}
}

//LCDにコマンドを出力する（4bit分割）
static	void	LcdCmd4bit(uint8_t data)
{
#if LCD_DB_REVERSE
	data = BitReverse(data);
#endif
	LCD_PORT &= ~LCD_DATABUS;
	LCD_PORT |= (data << LCD_DB_SHIFT) | LCD_E;
	_delay_ms(1);
	LCD_PORT &= ~LCD_E;
	_delay_ms(1);
}

//LCDにデータを出力する
void	LcdData(uint8_t data)
{
	LCD_PORT |= LCD_RS;
	
	LcdData4bit((data >> 4) & 0x0F);	//上位4bit
	LcdData4bit(data & 0x0F);			//下位4bit
	
	LCD_PORT &= ~LCD_RS;
	_delay_us(100);
}

//LCDにデータを出力する（4bit分割）
static	void	LcdData4bit(uint8_t data)
{
#if LCD_DB_REVERSE
	data = BitReverse(data);
#endif
	LCD_PORT &= ~LCD_DATABUS;
	LCD_PORT |= (data << LCD_DB_SHIFT) | LCD_E;	
	_delay_us(100);
	LCD_PORT &= ~LCD_E;
	_delay_us(100);
}

//LCD初期化
void	LCD_Init(uint8_t columnSize, uint8_t rowSize)
{
	LCD_DDR |= (LCD_DATABUS | LCD_E | LCD_RS);		//出力方向
	LCD_PORT &= ~(LCD_DATABUS | LCD_E | LCD_RS);	//Loを出力
	
	_delay_ms(15);
	LcdCmd4bit(0b0011);
	_delay_ms(5);
	LcdCmd4bit(0b0011);
	_delay_us(100);
	LcdCmd4bit(0b0011);
	
	LcdCmd4bit(0b0010);		//4-bitモードへ移行
	
	//   bit:76543210
	LcdCmd(0b00101000);		//Function Set	bit:5(1), DL=4-bit(0), N=2lines(1), F=5x8dots(0), *(0), *(0)
	LcdCmd(0b00001000);		//Display Set	bit:3(1), D=Set Display:NO(0), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
	LcdCmd(0b00000001);		//Clear Display	bit:0(1)
	LcdCmd(0b00000110);		//Entry Mode	bit:2(1), I/D=Increment(1), S=With Display Shift:NO(0)
	LcdCmd(0b00001100);		//Display Set	bit:3(1), D=Set Display:YES(1), C=Set Cursor:NO(0), B=Cursor Blink:NO(0)
	
	lcdInfo.cursorPosX = lcdInfo.cursorPosY = 0;
	lcdInfo.isCombine = lcdInfo.isOverwrite = 0;
	lcdInfo.XColumnMax = (columnSize < 1 ? 1 : columnSize);
	lcdInfo.YRowMax = (rowSize < 1 ? 1 : rowSize);
}

//カーソル表示状態
#define CMD_CursorOn	(LCD_CMD_DISPLAY_MODE | LCD_SUBCMD_SET_DISPLAY | LCD_SUBCMD_CURSOR_BLINK)
#define CMD_CursorOff	(LCD_CMD_DISPLAY_MODE | LCD_SUBCMD_SET_DISPLAY)

//「行」の動作モードをセットする
//・上書きモードの場合、どの位置まで表示が進んだか分からなくなるので、カーソルを表示できるようにする。
void	LCD_SetLineMode(uint8_t lineMode, uint8_t cursorFlag)
{
	lcdInfo.isCombine = lineMode & LCD_LINEMODE_COMBINE;
	lcdInfo.isOverwrite = lineMode & LCD_LINEMODE_OVERWRITE;
	LcdCmd(cursorFlag ? CMD_CursorOn : CMD_CursorOff);
}

//表示位置をセットする
//・行,桁とも0が起点。
void	LCD_SetPos(uint8_t row_y, uint8_t column_x)
{
	if (lcdInfo.XColumnMax <= column_x || lcdInfo.YRowMax <= row_y) { return; }
	
	LcdCmd(LCD_CMD_SET_DDRAM_ADRS | (row_y * 0x40 + column_x));
	lcdInfo.cursorPosX = column_x;
	lcdInfo.cursorPosY = row_y;
}

//LCD画面を消す
void	LCD_Clear(void)
{
	LcdCmd(LCD_CMD_CLEAR_DISPLAY);
}

//1文字出力する
void	LCD_PutChar(char c)
{
	if (lcdInfo.XColumnMax <= lcdInfo.cursorPosX || lcdInfo.YRowMax <= lcdInfo.cursorPosY) { return; }
	LcdData(c);
	lcdInfo.cursorPosX++;
	
	if (lcdInfo.cursorPosX == lcdInfo.XColumnMax)
	{
		if (lcdInfo.isCombine || lcdInfo.isOverwrite)
		{
			lcdInfo.cursorPosX = 0;
		}
		
		if (lcdInfo.isCombine)
		{
			lcdInfo.cursorPosY++;
			if (lcdInfo.cursorPosY == lcdInfo.YRowMax)
			{
				if (lcdInfo.isOverwrite)
				{
					lcdInfo.cursorPosY = 0;
				}
			}
		}
		
		LCD_SetPos(lcdInfo.cursorPosY, lcdInfo.cursorPosX);
	}
}

//文字列を表示する
void	LCD_PutString(const char* str)
{
	for (uint8_t i = 0; str[i] != '\0'; i++) { LCD_PutChar(str[i]); }
}

//プログラム領域に格納された文字列を表示する
void	LCD_PutStringP(const char* strpgm)
{
	uint8_t cc;
	while ((cc = pgm_read_byte(strpgm++)) != '\0') { LCD_PutChar(cc); }
}

//符号なし整数を出力する（10進数表記）
void	LCD_PutUInt(uint16_t n)
{
	int8_t	i = 0;
	uint8_t	kbuf[5];	//16bit正の整数は最大5桁(65535)
	
	do
	{
		kbuf[i++] = n % 10;
		n /= 10;
	}
	while (0 < n);
	
	while (i != 0) { LCD_PutChar('0' + kbuf[--i]); }
}

//符号あり整数を出力する（10進数表記）
void	LCD_PutInt(int16_t n)
{
	if (n < 0)
	{
		LCD_PutChar('-');
		LCD_PutUInt(-n);
	}
	else
	{
		LCD_PutUInt(n);
	}
}

//16進数を表示する（byte値／0詰め2桁）
void	LCD_PutByte(uint8_t n)
{
	uint8_t hex, cc;
	
	for (int8_t bitShift = 4; 0 <= bitShift; bitShift -= 4)
	{
		hex = (n >> bitShift) & 0x0F;
		cc = (hex < 10) ? ('0' + hex) : ('A' + hex - 10);
		LCD_PutChar(cc);
	}
}

//16進数を表示する（word値／0詰め4桁）
void	LCD_PutWord(uint16_t n)
{
	LCD_PutByte((n >> 8) & 0xFF);
	LCD_PutByte(n & 0xFF);
}

//ユーザー文字を定義する（詳細はHD44780のデータシートのコマンド一覧を参照）
//引数	id:		0-7（キャラクタID）
//		dots:	uint8_t[8]（5x8のドットパターン）
void	LCD_DefChar(uint8_t id, const uint8_t* dots)
{
	for (uint8_t i = 0; i < 8; i++)
	{
		LcdCmd(LCD_CMD_SET_CGRAM_ADRS | (id << 3) | i);
		LcdData(dots[i]);
	}
	
	//_delay_ms(30);	//LCDデバイス側の処理を待つ必要がある
}
